import pathlib

import safetensors.torch
import torch

from sd_mecha.sd_meh import utils as sd_meh_utils, merge as sd_meh_merge
from typing import Optional, Tuple


class MergeScheduler:
    def __init__(
        self, *,
        base_dir: Optional[pathlib.Path | str] = None,
        precision: int = 16,
        prune: bool = False,
        threads: int = 1,
        device: str = "cpu",
        work_device: Optional[str] = None,
    ):
        self.__base_dir = base_dir if base_dir is not None else base_dir
        if isinstance(self.__base_dir, str):
            self.__base_dir = pathlib.Path(self.__base_dir)
        self.__base_dir = self.__base_dir.absolute()

        self.__precision = precision
        self.__prune = prune
        self.__threads = threads
        self.__default_device = device
        self.__default_work_device = work_device

    def symbolic_merge(self, merge_method, a, b, c, alpha, beta, rebasin_iters, device, work_device, prune, threads):
        models = models_dict(a, b, c)
        weights, bases = weights_and_bases(merge_method, alpha, beta)

        return sd_meh_merge.merge_models(
            models,
            weights,
            bases,
            merge_method,
            self.__precision,
            False,
            bool(rebasin_iters),
            rebasin_iters if rebasin_iters is not None else 0,
            device if device is not None else self.__default_device,
            work_device if work_device is not None else self.__default_work_device,
            prune if prune is not None else self.__prune,
            threads if threads is not None else self.__threads,
        )

    def clip_weights(self, model, a, b, device):
        models = models_dict(a, b, None)
        return sd_meh_merge.clip_weights(models, model)

    def load_state_dict(self, path, device):
        if not isinstance(path, (str, pathlib.Path)):
            return path
        if not isinstance(path, pathlib.Path):
            path = pathlib.Path(path)
        if not path.suffix:
            path = path.with_suffix(".safetensors")
        if not path.is_absolute():
            path = self.__base_dir / path
        return sd_meh_merge.load_sd_model(path, device if device is not None else self.__default_device)

    def merge_and_save(self, merge_tree, *, output_path: Optional[pathlib.Path | str] = None):
        merged = merge_tree.visit(self)

        if not isinstance(output_path, pathlib.Path):
            output_path = pathlib.Path(output_path)
        if not output_path.is_absolute():
            output_path = self.__base_dir / output_path
        if not output_path.suffix:
            output_path = output_path.with_suffix(".safetensors")

        if output_path.suffix == ".safetensors":
            safetensors.torch.save_file(
                merged.to_dict(),
                f"{output_path}",
                metadata={"format": "pt"},
            )
        else:
            torch.save(
                {"state_dict": merged},
                f"{output_path.with_suffix('.ckpt')}",
            )


def models_dict(a, b, c=None) -> dict:
    models = {
        "model_a": a,
        "model_b": b,
    }
    if c is not None:
        models["model_c"] = c
    return models


def weights_and_bases(merge_method, alpha, beta=None) -> Tuple[dict, dict]:
    return sd_meh_utils.weights_and_bases(
        merge_method,
        None,
        alpha,
        None,
        None,
        beta,
        None,
        None,
        None,
        None,
        None,
    )



import pathlib
from typing import Optional

from sd_mecha.merge_scheduler import MergeScheduler
from sd_mecha import ast_nodes


def merge_and_save(
    merge_tree: ast_nodes.MergeNode,
    base_dir,
    output_path,
):
    scheduler = MergeScheduler(base_dir=base_dir)
    scheduler.merge_and_save(merge_tree, output_path=output_path)


def weighted_sum(
    a, b, *,
    alpha: float = 0.5,
    rebasin_iters: Optional[int] = None,
    precision: int = 16,
    prune: bool = False,
    threads: int = 1,
    device: str = "cpu",
    work_device: Optional[str] = None,
) -> ast_nodes.MergeNode:
    if isinstance(a, (str, pathlib.Path)):
        a = ast_nodes.LeafMergeNode(a, device)
    if isinstance(b, (str, pathlib.Path)):
        b = ast_nodes.LeafMergeNode(b, device)

    return ast_nodes.SymbolicMergeNode(
        merge_method="weighted_sum",
        a=a,
        b=b,
        alpha=alpha,
        rebasin_iters=rebasin_iters,
        precision=precision,
        prune=prune,
        threads=threads,
        device=device,
        work_device=work_device,
    )


def add_difference(
    a, b, c, *,
    alpha: float = 1.0,
    precision: int = 16,
    prune: bool = False,
    threads: int = 1,
    device: str = "cpu",
    work_device: Optional[str] = None,
) -> ast_nodes.MergeNode:
    if isinstance(a, (str, pathlib.Path)):
        a = ast_nodes.LeafMergeNode(a, device)
    if isinstance(b, (str, pathlib.Path)):
        b = ast_nodes.LeafMergeNode(b, device)
    if isinstance(c, (str, pathlib.Path)):
        c = ast_nodes.LeafMergeNode(c, device)

    return ast_nodes.SymbolicMergeNode(
        merge_method="add_difference",
        a=a,
        b=b,
        c=c,
        alpha=alpha,
        precision=precision,
        prune=prune,
        threads=threads,
        device=device,
        work_device=work_device,
    )


def rotate(
    a, b, *,
    alpha: float = 1.0,
    beta: float = 0.0,
    precision: int = 16,
    prune: bool = False,
    threads: int = 1,
    device: str = "cpu",
    work_device: Optional[str] = None,
) -> ast_nodes.MergeNode:
    if isinstance(a, (str, pathlib.Path)):
        a = ast_nodes.LeafMergeNode(a, device)
    if isinstance(b, (str, pathlib.Path)):
        b = ast_nodes.LeafMergeNode(b, device)

    return ast_nodes.SymbolicMergeNode(
        merge_method="add_difference",
        a=a,
        b=b,
        alpha=alpha,
        beta=beta,
        precision=precision,
        prune=prune,
        threads=threads,
        device=device,
        work_device=work_device,
    )


def add_perpendicular(
    a, b, c, *,
    alpha: float = 1.0,
    rebasin_iters: Optional[int] = None,
    precision: int = 16,
    prune: bool = False,
    threads: int = 1,
    device: str = "cpu",
    work_device: Optional[str] = None,
) -> ast_nodes.MergeNode:
    if isinstance(a, (str, pathlib.Path)):
        a = ast_nodes.LeafMergeNode(a, device)
    if isinstance(b, (str, pathlib.Path)):
        b = ast_nodes.LeafMergeNode(b, device)
    if isinstance(c, (str, pathlib.Path)):
        c = ast_nodes.LeafMergeNode(c, device)

    return ast_nodes.SymbolicMergeNode(
        merge_method="add_perpendicular",
        a=a,
        b=b,
        c=c,
        alpha=alpha,
        rebasin_iters=rebasin_iters,
        precision=precision,
        prune=prune,
        threads=threads,
        device=device,
        work_device=work_device,
    )


def clip(model, a, b, device: str = None) -> ast_nodes.MergeNode:
    if isinstance(a, (str, pathlib.Path)):
        a = ast_nodes.LeafMergeNode(a, device)
    if isinstance(b, (str, pathlib.Path)):
        b = ast_nodes.LeafMergeNode(b, device)

    return ast_nodes.ClipMergeNode(model, a, b, device)



import abc
from typing import Optional


class MergeNode(abc.ABC):
    @abc.abstractmethod
    def visit(self, scheduler):
        pass


class LeafMergeNode(MergeNode):
    def __init__(self, state_dict, device=None):
        self.__state_dict = state_dict
        self.__device = device

    def visit(self, scheduler):
        return scheduler.load_state_dict(self.__state_dict, self.__device)


class SymbolicMergeNode(MergeNode):
    def __init__(
        self,
        merge_method,
        a,
        b,
        alpha,
        c=None,
        beta=None,
        rebasin_iters=None,
        precision: int = 16,
        prune: bool = False,
        threads: int = 1,
        device: str = "cpu",
        work_device: Optional[str] = None,
    ):
        self.__merge_method = merge_method
        self.__a = a
        self.__b = b
        self.__c = c
        self.__alpha = alpha
        self.__beta = beta
        self.__rebasin_iters = rebasin_iters
        self.__precision = precision
        self.__prune = prune
        self.__threads = threads
        self.__device = device
        self.__work_device = work_device

    def visit(self, scheduler):
        return scheduler.symbolic_merge(
            self.__merge_method,
            self.__a.visit(scheduler),
            self.__b.visit(scheduler),
            self.__c.visit(scheduler) if self.__c is not None else None,
            self.__alpha, self.__beta,
            self.__rebasin_iters,
            self.__device, self.__work_device,
            self.__prune,
            self.__threads,
        )


class ClipMergeNode(MergeNode):
    def __init__(self, model, a, b, device=None):
        self.__model = model
        self.__a = a
        self.__b = b
        self.__device = device

    def visit(self, scheduler):
        return scheduler.clip_weights(self.__model, self.__a, self.__b, self.__device)
